import sys
from .conditions import Condition, AltCondition, process_folder, process_alt_folder
from ._version import __version__

if sys.version_info[0] < 3 and sys.version_info[1] < 4:
    raise Exception('Python 3.4 or greater is required to use this package.')



"""Simulations Module"""

# Third party imports
import numpy as np
import cantera as ct


class Simulation(object):
    """Contains a single simulation of the experiment.

    Parameters
    ----------
    initial_temperature : `float`
        The initial temperature of the simulation
    initial_pressure : `float`
        The initial pressure of the simulation
    volume : `numpy.ndarray` or `None`
        The volume trace to be used for the simulation. Must be
        supplied, but if the input value is `None`, the volume trace
        will be read from the file ``volume.csv``. The first column
        should be the time, the second column should be the volume.
    is_reactive : `bool`
        If the simulation should be reactive or non-reactive. If `False`
        sets the Cantera reaction rate multiplier to 0.0 via the
        `~cantera.Kinetics.set_multiplier` function.
    end_temp : `float`, optional
        Reactor temperature at which the simulation will be ended
    end_time : `float`, optional
        Time at which the simulation will be ended
    chem_file : `str`, optional
        String filename of the chemistry file to use

    Attributes
    ----------
    time : `numpy.ndarray`
        Array of simulated time values
    temperature : `numpy.ndarray`
        Array of simulated temperature values
    pressure : `numpy.ndarray`
        Array of simulated pressure values
    input_volume : `numpy.ndarray`
        Array of input volume values
    simulated_volume : `numpy.ndarray`
        Array of simulated volume values
    end_temp : `float`
        Reactor temperature at which the simulation will be ended
    end_time : `float`
        Time at which the simulation will be ended
    chem_file : `str`
        String filename of the chemistry file to use
    initial_temperature : `float`
        The initial temperature of the simulation
    initial_pressure : `float`
        The initial pressure of the simulation
    """

    def __init__(self, initial_temperature, initial_pressure, volume, is_reactive,
                 end_temp=2500., end_time=0.2, chem_file='species.cti', cti_source=None):

        if volume is None:
            volume = np.genfromtxt('volume.csv', delimiter=',')

        inp_time = volume[:, 0]
        inp_vol = volume[:, 1]

        self.time = []
        self.temperature = []
        self.pressure = []
        self.input_volume = volume
        self.simulated_volume = []
        self.end_temp = end_temp
        self.end_time = end_time
        self.is_reactive = is_reactive
        self.chem_file = chem_file
        self.initial_temperature = initial_temperature
        self.initial_pressure = initial_pressure

        if cti_source is None:
            gas = ct.Solution(chem_file)
        else:
            gas = ct.Solution(source=cti_source)
        gas.TP = self.initial_temperature, self.initial_pressure
        if not self.is_reactive:
            gas.set_multiplier(0)
        reac = ct.IdealGasReactor(gas)
        env = ct.Reservoir(ct.Solution('air.xml'))
        ct.Wall(reac, env, A=1.0, velocity=VolumeProfile(inp_time, inp_vol))
        netw = ct.ReactorNet([reac])
        netw.set_max_time_step(inp_time[1])
        self.time.append(netw.time)
        self.temperature.append(reac.T)
        self.pressure.append(gas.P/1E5)
        self.simulated_volume.append(reac.volume)

        while reac.T < self.end_temp and netw.time < self.end_time:
            netw.step()
            self.time.append(netw.time)
            self.temperature.append(reac.T)
            self.pressure.append(gas.P/1E5)
            self.simulated_volume.append(reac.volume)

        self.time = np.array(self.time)
        self.pressure = np.array(self.pressure)
        self.temperature = np.array(self.temperature)
        self.simulated_volume = np.array(self.simulated_volume)
        self.derivative = self.calculate_derivative(self.pressure, self.time)

    def __repr__(self):
        return ('Simulation(initial_temperature={self.initial_temperature!r}, '
                'initial_pressure={self.initial_pressure!r}, volume={self.input_volume!r}, '
                'is_reactive={self.is_reactive!r}, end_temp={self.end_temp!r}, '
                'end_time={self.end_time!r}, chem_file={self.chem_file!r})').format(
                    self=self,
                )

    def calculate_derivative(self, dep_var, indep_var):
        """Calculate the derivative.

        Parameters
        ----------
        dep_var : `numpy.ndarray`
            Dependent variable (e.g., the pressure)
        indep_var : `numpy.ndarray`
            Independent variable (e.g., the time)

        Returns
        -------
        `numpy.ndarray`
            1-D array containing the derivative

        Notes
        -----
        The derivative is calculated by computing the first-order
        Lagrange polynomial fit to the point under consideration and
        its nearest neighbors. The Lagrange polynomial is used because
        of the unequal spacing of the simulated data.
        """
        m = len(dep_var)
        ddt = np.zeros(m)
        for i in range(1, m-2):
            x = indep_var[i]
            x_min = indep_var[i-1]
            x_plu = indep_var[i+1]
            y = dep_var[i]
            y_min = dep_var[i-1]
            y_plu = dep_var[i+1]
            ddt[i] = (y_min*(x - x_plu)/((x_min - x)*(x_min - x_plu)) +
                      y*(2*x - x_min - x_plu)/((x - x_min)*(x - x_plu)) +
                      y_plu*(x - x_min)/((x_plu - x_min)*(x_plu - x)))

        return ddt


class VolumeProfile(object):
    """
    Set the velocity of the piston by using a user specified volume
    profile. The initialization and calling of this class are handled
    by the `cantera.Func1` interface of Cantera.

    The velocity is calculated by assuming a unit area and using the
    forward difference, calculated by `numpy.diff`. This function is
    only called once when the class is initialized at the beginning of
    a problem so it is efficient.

    Parameters
    ----------
    time: `numpy.ndarray`
        Array or list of time values
    volume: `numpy.ndarray`
        Array or list of volume values

    Attributes
    ----------
    time: `numpy.ndarray`
        Array of time values
    volume: `numpy.ndarray`
        Array of volume values
    velocity: `numpy.ndarray`
        Array of velocity values
    """

    def __init__(self, time, volume):
        # The time and volume are stored as lists in the keywords
        # dictionary. The volume is normalized by the first volume
        # element so that a unit area can be used to calculate the
        # velocity.
        self.time = np.array(time)
        self.volume = np.array(volume)/volume[0]

        # The velocity is calculated by the forward difference.
        # numpy.diff returns an array one element smaller than the
        # input array, so we append a zero to match the length of the
        # self.time array.
        self.velocity = np.diff(self.volume)/np.diff(self.time)
        self.velocity = np.append(self.velocity, 0)

    def __repr__(self):
        return ('VolumeProfile(time={self.time!r}, volume={self.volume!r}, '
                'velocity={self.velocity!r}').format(
                    self=self,
                )

    def __call__(self, t):
        """Return the velocity when called during a time step.

        Parameters
        ----------
        t : `float`
            Current simulation time.
        """

        if t < self.time[-1]:
            # prev_time_point is the previous value in the time array
            # after the current simulation time
            prev_time_point = self.time[self.time <= t][-1]
            # index is the index of the time array where
            # prev_time_point occurs
            index = np.where(self.time == prev_time_point)[0][0]
            return self.velocity[index]
        else:
            return 0



"""All of the kinds of traces in UConnRCMPy"""

# System imports

# Third-party imports
import numpy as np
import cantera as ct
from scipy import signal as sig
from scipy.interpolate import UnivariateSpline
from scipy.stats import linregress

# Local imports
from .constants import (one_atm_in_bar,
                        one_atm_in_torr,
                        one_bar_in_pa,
                        )


class VoltageTrace(object):
    """Voltage signal from a single experiment.

    Parameters
    ----------
    file_path : `pathlib.Path`
        `~pathlib.Path` object associated with the particular experiment

    Attributes
    ----------
    signal : `numpy.ndarray`
        2-D array containing the raw signal from the experimental
        text file. First column is the time, second column is the
        voltage.
    time : `numpy.ndarray`
        The time loaded from the signal trace
    frequency : `int`
        The sampling frequency of the pressure trace
    filtered_voltage : `numpy.ndarray`
        The voltage trace after filtering

    Note
    ----
    The first sample of the voltage is set equal to the
    mean of the first 200 points to eliminate DAQ startup
    effects seen in some data.
    """
    def __init__(self, file_path):
        self.file_path = file_path
        self.signal = np.genfromtxt(str(self.file_path))
        self.time = self.signal[:, 0]
        self.frequency = np.rint(1/self.time[1])

        self.filter_frequency = None

        self.signal[0, 1] = np.mean(self.signal[:200, 1])

        self.filtered_voltage = self.filtering(self.signal[:, 1])

    def __repr__(self):
        return 'VoltageTrace(file_path={self.file_path!r})'.format(self=self)

    @property
    def filter_frequency(self):
        """The cutoff frequency for the low-pass filter

        When setting the frequency, if the ``value`` is `None`,
        determines the optimal cutoff frequency for a first-order
        Butterworth low-pass filter by analyzing the root-mean-squared
        residuals for a sequence of cutoff frequencies. The residuals
        plotted as a function of the cutoff frequency tend to have a
        linear portion for a range of cutoff frequencies. Analysis of
        typical data files from our RCM has shown this range to start
        near ``nyquist_freq*0.05``. The end point is determined by
        looping through values from ``nyquist_freq*0.5`` to
        ``nyquist_freq*0.1`` and finding the location where the
        coefficient of determination of a linear fit is maximized.
        A line is fit to this portion of the residuals curve and
        the intersection point of a horizontal line through the
        y-intercept of the fit and the residuals curve is used to
        determine the optimal cutoff frequency (see Figure 2 in Yu
        et al. [1]_). The methodology is described by Yu et al. [1]_,
        and the code is modifed from Duarte [2]_.

        References
        ----------
        .. [1] B. Yu, D. Gabriel, L. Noble, and K.N. An, "Estimate of
               the Optimum Cutoff Frequency for the Butterworth Low-Pass
               Digital Filter", Journal of Applied Biomechanics, Vol. 15,
               pp. 318-329, 1999.
               DOI: `10.1123/jab.15.3.318 <http://dx.doi.org/10.1123/jab.15.3.318>`_

        .. [2] M. Duarte, "Residual Analysis", v.3 2014/06/13,
               http://nbviewer.ipython.org/github/demotu/BMC/blob/master/notebooks/ResidualAnalysis.ipynb
        """

        return self._filter_frequency

    @filter_frequency.setter
    def filter_frequency(self, value):
        if value is None:
            nyquist_freq = self.frequency/2.0
            n_freqs = 101
            freqs = np.linspace(nyquist_freq/n_freqs, nyquist_freq, n_freqs)
            resid = np.zeros(n_freqs)
            for i, fc in enumerate(freqs):
                b, a = sig.butter(1, fc/nyquist_freq)
                yf = sig.filtfilt(b, a, self.signal[:, 1])
                resid[i] = np.sqrt(np.mean((yf - self.signal[:, 1])**2))

            end_points = np.linspace(0.5, 0.1, 9)
            r_sq = np.zeros(len(end_points))
            intercepts = np.zeros(len(end_points))
            for i, end_point in enumerate(end_points):
                # The indices of the frequencies used for fitting the straight line
                fit_freqs = np.arange(np.nonzero(freqs >= nyquist_freq*0.05)[0][0],
                                      np.nonzero(freqs >= nyquist_freq*end_point)[0][0] + 1)
                _, intercepts[i], r, _, _ = linregress(freqs[fit_freqs], resid[fit_freqs])
                r_sq[i] = r**2

            intercept = intercepts[np.argmax(r_sq)]

            # The UnivariateSpline with s=0 forces the spline fit through every
            # data point in the array. The residuals are shifted down by the
            # intercept so that the root of the spline is the optimum cutoff
            # frequency
            try:
                self._filter_frequency = UnivariateSpline(freqs, resid - intercept, s=0).roots()[0]
            except IndexError:
                self._filter_frequency = float(input(
                    'Automatic setting of the filter frequency failed. Please input a frequency; '
                    'typical values are between 1000-5000 Hz: '))
        else:
            self._filter_frequency = value

    def change_filter_freq(self, value):
        """Change the filter frequency

        This method is intended to be used after the `VoltageTrace` has
        already been created. It sets the `~VoltageTrace.filter_frequency`
        attribute and then runs the filtering.

        Parameters
        ----------
        value : `float` or `None`
            Value for the filter frequency. If `None`, an optimal frequency
            will be estimated by the procedure detailed in the documentation
            for the `~VoltageTrace.filter_frequency` attribute.
        """
        self.filter_frequency = value
        self.filtered_voltage = self.filtering(self.signal[:, 1])

    def savetxt(self, filename, **kwargs):
        """Save a text file output of the voltage trace.

        Save a text file with the time in the first column and the filtered
        voltage in the second column. The keyword arguments are the same as
        `numpy.savetxt`.

        Parameters
        ----------
        filename : `str`
            Filename of the output file
        """
        np.savetxt(fname=filename, X=np.vstack(self.time, self.filtered_voltage).T, **kwargs)

    def filtering(self, data):
        """Filter the input using a low-pass filter.

        The filter is a first-order Butterworth filter, with the
        cutoff frequency taken from the instance attribute
        `~VoltageTrace.filter_frequency`.

        Parameters
        ----------
        data : `numpy.ndarray`
            The data that should be filtered

        Returns
        -------
        `numpy.ndarray`
            1-D array of the same length as the input data
        """
        nyquist_freq = self.frequency/2.0
        b, a = sig.butter(1, (self.filter_frequency)/nyquist_freq)
        return sig.filtfilt(b, a, data, padtype='odd', padlen=101, method='pad')


class ExperimentalPressureTrace(object):
    """Pressure trace from a single experiment.

    Parameters
    ----------
    voltage_trace : `VoltageTrace`
        Instance of class containing the voltage trace of the
        experiment.
    initial_pressure_in_torr : `float`
        The initial pressure of the experiment, in units of Torr
    factor : `float`
        The factor set on the charge amplifier

    Attributes
    ----------
    pressure : `numpy.ndarray`
        The pressure trace computed from the filtered voltage trace
    time : `numpy.ndarray`
        A 1-D array containting the time. Copied from
        `VoltageTrace.time`
    frequency : `int`
        Integer sampling frequency of the experiment. Copied from
        `VoltageTrace.frequency`
    p_EOC : `float`
        Pressure at the end of compression
    EOC_idx : `int`
        Integer index in the `pressure` and `time` arrays
        of the end of compression.
    is_reactive : `bool`
        Boolean if the pressure trace represents a reactive or
        or non-reactive experiment
    derivative : `numpy.ndarray`
        1-D array containing the raw derivative computed from the
        `pressure` trace.
    zeroed_time : `numpy.ndarray`
        1-D array containing the time, with the zero point set at
        the end of compression.
    """
    def __init__(self, voltage_trace, initial_pressure_in_torr, factor):
        initial_pressure_in_bar = initial_pressure_in_torr*one_atm_in_bar/one_atm_in_torr
        self.pressure = (voltage_trace.filtered_voltage - voltage_trace.filtered_voltage[0])
        self.pressure *= factor
        self.pressure += initial_pressure_in_bar

        self.raw_pressure = (voltage_trace.signal[:, 1] - voltage_trace.signal[0, 1])
        self.raw_pressure *= factor
        self.raw_pressure += initial_pressure_in_bar

        self.time = voltage_trace.time
        self.frequency = voltage_trace.frequency

        self.p_EOC, self.EOC_idx, self.is_reactive = self.find_EOC()
        self.derivative = self.calculate_derivative(self.pressure, self.time)
        # Smooth the derivative with a moving average 151 points wide
        self.derivative = sig.fftconvolve(self.derivative, np.ones(151)/151, mode='same')
        self.zeroed_time = self.time - self.time[self.EOC_idx]

    def __repr__(self):
        return ('ExperimentalPressureTrace(p_EOC={self.p_EOC!r}, '
                'is_reactive={self.is_reactive!r})').format(self=self)

    def savetxt(self, filename, **kwargs):
        """Save a text file output of the pressure trace.

        Save a text file with the time in the first column and the filtered
        pressure in the second column. The keyword arguments are the same as
        `numpy.savetxt`.

        Parameters
        ----------
        filename : `str`
            Filename of the output file
        """
        np.savetxt(fname=filename, X=np.vstack(self.time, self.pressure).T, **kwargs)

    def pressure_fit(self, comptime=0.08):
        """Fit a line to the pressure trace before compression starts.

        Parameters
        ----------
        comptime : `float`, optional
            Desired compression time, computed from the EOC, to when
            the pressure fit should start

        Returns
        -------
        `numpy.polyfit`
            Numpy object containing the parameters of the fit
        """
        beg_compress = int(np.floor(self.EOC_idx - comptime*self.frequency))
        time = np.linspace(0, (beg_compress - 1)/self.frequency, beg_compress)
        fit_pres = self.pressure[:beg_compress]
        fit_pres[0:9] = fit_pres[10]
        linear_fit = np.polyfit(time, fit_pres, 1)
        return linear_fit

    def change_EOC_time(self, time, is_reactive=True):
        """Change the EOC time for an experiment

        Parameters
        ----------
        time : `float`
            The new value of the EOC time
        is_reactive : `boolean`
            The experiment is reactive or not
        """
        offset = int(round(time/1000*self.frequency, 0))
        self.EOC_idx += offset
        self.p_EOC = self.pressure[self.EOC_idx]
        self.is_reactive = is_reactive
        self.zeroed_time = self.time - self.time[self.EOC_idx]

    def find_EOC(self):
        """Find the index and pressure at the end of compression.

        Returns
        -------
        `tuple`
            Returns a tuple with types (`float`, `int`,
            `bool`) representing the pressure at the end of
            compression, the index of the end of compression relative
            to the start of the pressure trace, and a boolean that is
            True if the case is reactive and False otherwise,
            respectively

        Notes
        -----
        The EOC is found by moving backwards from the maximum pressure
        point and testing the values of the pressure. When the test
        value becomes greater than the previous pressure, we have reached
        the minimum pressure before ignition, in the case of a reactive
        experiment. Then, the EOC is the maximum of the pressure before
        this minimum point. If the pressure at the minimum is close to
        the initial pressure, assume the case is non-reactive and set
        the EOC pressure and the index to the max pressure point.
        """
        is_reactive = True
        max_p = np.amax(self.pressure)
        max_p_idx = np.argmax(self.pressure)
        min_p_idx = max_p_idx - 100
        while self.pressure[min_p_idx] >= self.pressure[min_p_idx - 50]:
            min_p_idx -= 1

        p_EOC = np.amax(self.pressure[0:min_p_idx])
        p_EOC_idx = np.argmax(self.pressure[0:min_p_idx])
        diff = abs(self.pressure[p_EOC_idx] - self.pressure[15])
        if diff < 5.0:
            p_EOC, p_EOC_idx = max_p, max_p_idx
            is_reactive = False

        return p_EOC, p_EOC_idx, is_reactive

    def calculate_derivative(self, dep_var, indep_var):
        """Calculate the derivative.

        Parameters
        ----------
        dep_var : `numpy.ndarray`
            Dependent variable (e.g., the pressure)
        indep_var : `numpy.ndarray`
            Independent variable (e.g., the time)

        Returns
        -------
        `numpy.ndarray`
            1-D array containing the derivative

        Notes
        -----
        The derivative is calculated by a second-order forward method
        and any places where the derivative is infinite are set to
        zero.
        """
        m = len(dep_var)
        ddt = np.zeros(m)
        ddt[:m-2] = (-3*dep_var[:m-2] + 4*(dep_var[1:m-1]) - dep_var[2:m])/(2*np.diff(indep_var[:m-1]))  # NOQA
        ddt[np.isinf(ddt)] = 0
        return ddt


class AltExperimentalPressureTrace(ExperimentalPressureTrace):
    """Process an alternate experimental pressure trace.

    These pressure traces do not have an associated voltage trace,
    but the machinery in the VoltageTrace class is useful for
    filtering.
    """
    def __init__(self, file_path, initial_pressure_in_torr):
        # This is not a real voltage trace
        pressure_trace = VoltageTrace(file_path)

        self.time = pressure_trace.time
        self.frequency = pressure_trace.frequency
        self.filter_frequency = pressure_trace.filter_frequency

        self.pressure = pressure_trace.filtered_voltage
        pressure_start = np.mean(self.pressure[20:500])
        self.pressure -= pressure_start
        self.pressure += initial_pressure_in_torr*one_atm_in_bar/one_atm_in_torr

        self.raw_pressure = pressure_trace.signal[:, 1]
        raw_pressure_start = np.mean(self.raw_pressure[20:500])
        self.raw_pressure -= raw_pressure_start
        self.raw_pressure += initial_pressure_in_torr*one_atm_in_bar/one_atm_in_torr

        self.p_EOC, self.EOC_idx, self.is_reactive = self.find_EOC()
        self.derivative = self.calculate_derivative(self.pressure, self.time)
        self.zeroed_time = self.time - self.time[self.EOC_idx]

    def __repr__(self):
        return ('AltExperimentalPressureTrace(p_EOC={self.p_EOC!r}, '
                'is_reactive={self.is_reactive!r})').format(self=self)


class PressureFromVolume(object):
    """Create a pressure trace given a volume trace.

    Using Cantera to evaluate the thermodynamic properties, compute a
    pressure trace from a volume trace.

    Parameters
    ----------
    volume : `numpy.ndarray`
        1-D array containing the reactor volume
    p_initial : `float`
        Initial pressure of the experiment, in bar
    T_initial : `float`, optional
        Initial temperature of the experiment, in Kelvin.
        Optional for Cantera versions greater than 2.2.0.
    chem_file : `str`, optional
        Filename of the chemistry file to be used

    Attributes
    ----------
    pressure : `numpy.ndarray`
        The pressure trace

    Notes
    -----
    The pressure is computed in a `cantera.Solution` object by
    setting the volume and the entropy according to an isentropic
    process using the given volume trace.
    """
    def __init__(self, volume, p_initial, T_initial, chem_file='species.cti', cti_source=None):
        if cti_source is None:
            gas = ct.Solution(chem_file)
        else:
            gas = ct.Solution(source=cti_source)
        gas.TP = T_initial, p_initial
        initial_volume = gas.volume_mass
        initial_entropy = gas.entropy_mass
        self.pressure = np.zeros((len(volume)))
        for i, v in enumerate(volume):
            gas.SV = initial_entropy, v*initial_volume
            self.pressure[i] = gas.P/one_bar_in_pa

    def __repr__(self):
        return 'PressureFromVolume(pressure={self.pressure!r})'.format(self=self)


class VolumeFromPressure(object):
    r"""Create a volume trace given a pressure trace.

    Using Cantera to evaluate the thermodynamic properties, compute a
    volume trace from a pressure trace.

    Parameters
    ----------
    pressure : `numpy.ndarray`
        1-D array containing the reactor pressure
    v_initial : `float`
        Initial volume of the experiment, in m**3
    T_initial : `float`, optional
        Initial temperature of the experiment, in Kelvin. Optional for
        Cantera versions greater than 2.2.0.
    chem_file : `str`, optional
        Filename of the chemistry file to be used

    Attributes
    ----------
    volume : `numpy.ndarray`
        The volume trace

    Notes
    -----
    The volume is computed according to the formula

    .. math:: v_i = v_{initial}*\rho_{initial}/\rho_i

    where the index :math:`i` indicates the current point. The state
    is set at each point by setting the pressure from the input array
    and the entropy to be constant. The volume is computed by the
    isentropic relationship described above.
    """
    def __init__(self, pressure, v_initial, T_initial, chem_file='species.cti', cti_source=None):
        if cti_source is None:
            gas = ct.Solution(chem_file)
        else:
            gas = ct.Solution(source=cti_source)
        gas.TP = T_initial, pressure[0]*one_bar_in_pa
        initial_entropy = gas.entropy_mass
        initial_density = gas.density
        self.volume = np.zeros((len(pressure)))
        for i, p in enumerate(pressure):
            gas.SP = initial_entropy, p*one_bar_in_pa
            self.volume[i] = v_initial*initial_density/gas.density

    def __repr__(self):
        return 'VolumeFromPressure(volume={self.volume!r})'.format(self=self)


class TemperatureFromPressure(object):
    """Create a temperature trace given a pressure trace.

    Using Cantera to evaluate the thermodynamic properties, compute a
    pressure trace from a volume trace.

    Parameters
    ----------
    pressure : `numpy.ndarray`
        1-D array containing the pressure
    T_initial : `float`
        Initial temperature of the experiment, in Kelvin.
        Optional for Cantera versions greater than 2.2.0.
    chem_file : `str`, optional
        Filename of the chemistry file to be used

    Attributes
    ----------
    temperature : `numpy.ndarray`
        The temperature trace

    Notes
    -----
    The temperature is computed in a `cantera.Solution` object by
    setting the pressure and the entropy according to an isentropic
    process using the given pressure trace.
    """
    def __init__(self, pressure, T_initial, chem_file='species.cti', cti_source=None):
        if cti_source is None:
            gas = ct.Solution(chem_file)
        else:
            gas = ct.Solution(source=cti_source)
        gas.TP = T_initial, pressure[0]*one_bar_in_pa
        initial_entropy = gas.entropy_mass
        self.temperature = np.zeros((len(pressure)))
        for i, p in enumerate(pressure):
            gas.SP = initial_entropy, p*one_bar_in_pa
            self.temperature[i] = gas.T

    def __repr__(self):
        return 'TemperatureFromPressure(temperature={self.temperature!r})'.format(self=self)



"""Experiments module"""

# System imports
from datetime import datetime
from pathlib import Path
import platform

# Third-part imports
import cantera as ct
import numpy as np
import matplotlib.pyplot as plt

# Local imports
from pyperclip import copy
from .traces import (VoltageTrace,
                     ExperimentalPressureTrace,
                     AltExperimentalPressureTrace,
                     TemperatureFromPressure,
                     )


class Experiment(object):
    """Contains all the information of a single RCM experiment.

    Parameters
    ----------
    file_path : `str` or `pathlib.Path`, optional
        If an argument is supplied, it should be an instance of
        `~pathlib.Path` or `str`. If no argument is supplied, the
        filename is read from the standard input as a string and
        resolved into a `~pathlib.Path` object.
    cti_file : `str` or `pathlib.Path`, optional
        Location of the CTI file for Cantera
    cti_source : `str`, optional
        String containing the source of a CTI file for Cantera
    copy : `bool`, optional
        Boolean indicating whether values should be copied to
        the clipboard.

    Attributes
    ----------
    compression_time : `float`
        The compression time in milliseconds, from the end of compression to the
        approximate start of piston motion
    file_path : `pathlib.Path`
        Object storing the file path
    experiment_parameters : `dict`
        Stores the parameters of the experiment parsed from the
        filename by `~uconnrcmpy.Experiment.parse_file_name`.
    voltage_trace : `~uconnrcmpy.traces.VoltageTrace`
        Stores the experimental voltage signal and related traces
    pressure_trace : `~uconnrcmpy.traces.ExperimentalPressureTrace`
        Stores the experimental pressure trace and its derivative
    output_end_time : `float`
        The end time for the output to a file, relative to the end of compression in milliseconds
    offset_points : `float`
        The number of points to offset the end of compression, to better
        match the non-reactive and reactive pressure traces together
    ignition_delay : `float`
        The overall ignition delay of the experiment. Will be zero for
        a non-reactive experiment.
    first_stage : `float`
        The first stage ignition delay of the experiment. Will be zero
        for a non-reactive experiment of if there is no first-stage
        ignition.
    T_EOC : `float`
        The temperature estimated at the end of compression
    """

    def __init__(self, file_path=None, cti_file=None, cti_source=None, copy=True):
        self.resolve_file_path(file_path)
        self.experiment_parameters = self.parse_file_name(self.file_path)
        self.voltage_trace = VoltageTrace(self.file_path)
        self.pressure_trace = ExperimentalPressureTrace(self.voltage_trace,
                                                        self.experiment_parameters['pin'],
                                                        self.experiment_parameters['factor'],
                                                        )
        self.compression_time = None
        self.output_end_time = None
        self.offset_points = None
        if cti_source is None and cti_file is None:
            raise ValueError('One of cti_file or cti_source must be specified')
        elif cti_source is not None and cti_file is not None:
            raise ValueError('Only one of cti_file or cti_source can be specified')
        elif cti_source is None:
            self.cti_file = Path(cti_file).resolve()
            with open(str(cti_file), 'r') as in_file:
                self.cti_source = in_file.read()
        else:
            self.cti_source = cti_source
        self.process_pressure_trace()
        if copy:
            self.copy_to_clipboard()

    def __repr__(self):
        return 'Experiment(file_path={self.file_path!r})'.format(self=self)

    def parse_file_name(self, file_path):
        """Parse the file name of an experimental trace.

        Parameters
        ----------
        file_path : :class:`pathlib.Path`
            The Path object that contains the path of the current experimental data file.
            The filename associated with the path should be in the following format::

                [NR_]XX_in_YY_mm_ZZZK-AAAAt-BBBx-DD-Mon-YY-Time.txt

            where:

            - ``[NR_]``: Optional non-reactive indicator
            - ``XX``: inches of spacers
            - ``YY``: millimeters of shims
            - ``ZZZ``: Initial temperature in Kelvin
            - ``AAAA``: Initial pressure in Torr
            - ``BBB``: Multiplication factor set on the charge amplifier
            - ``DD-Mon-YY-Time``: Day, Month, Year, and Time of experiment

        Returns
        -------
        :class:`dict`
            Dictionary containing the parameters of the experiment with the
            following names:

            - ``spacers``: Inches of spacers
            - ``shims``: Millimeters of shims
            - ``Tin``: Initial temperature in Kelvin
            - ``pin``: Initial pressure in Torr
            - ``factor``: Multiplication factor set on the charge amplifier
            - ``time_of_day``: Time of day of the experiment
            - ``date``: Date of the experiment with the time
            - ``date_year``: Date of the experiment with the year
        """
        name_parts = {}
        fname = file_path.name.lstrip('NR_')
        fname = fname.rstrip('.txt')
        name_split = fname.split('_')
        name_split_end = name_split[4].split('-', maxsplit=3)
        data_date = datetime.strptime(name_split_end[3], '%d-%b-%y-%H%M')

        name_parts['spacers'] = int(name_split[0])/10
        name_parts['shims'] = int(name_split[2])
        name_parts['Tin'] = int(name_split_end[0][:-1])
        name_parts['pin'] = int(name_split_end[1][:-1])
        name_parts['factor'] = int(name_split_end[2][:-1])
        name_parts['time_of_day'] = data_date.strftime('%H%M')
        name_parts['date'] = data_date.strftime('%d-%b-%H:%M')
        name_parts['date_year'] = data_date.strftime('%d-%b-%y')
        return name_parts

    def resolve_file_path(self, file_path=None):
        if file_path is None:
            self.file_path = Path(input('Filename: '))
        else:
            self.file_path = Path(file_path)
        try:
            self.file_path = self.file_path.resolve()
        except FileNotFoundError:
            self.file_path = self.file_path.with_suffix('.txt')
            self.file_path = self.file_path.resolve()

    def process_pressure_trace(self):
        if self.pressure_trace.is_reactive:
            self.ignition_delay, self.first_stage = self.calculate_ignition_delay()

            try:
                self.T_EOC = self.calculate_EOC_temperature()
            except ct.CanteraError as e:
                self.T_EOC = 0
                print('Exception in computing the temperature at EOC', e)
        else:
            self.ignition_delay, self.first_stage, self.T_EOC = 0, 0, 0

    def change_EOC_time(self, time, is_reactive=True):
        """Change the EOC time for an experiment

        Parameters
        ----------
        time : `float`
            The new value of the EOC time
        is_reactive : `boolean`
            The experiment is reactive or not
        """
        self.pressure_trace.change_EOC_time(time, is_reactive)
        self.process_pressure_trace()
        self.copy_to_clipboard()
        if hasattr(self, 'p_axis'):
            # Create a list of the lines in the axis so that
            # removing one doesn't affect the list as its
            # looping. Use the loop rather than axis.clear()
            # so that the axis labels and legend are preserved.
            for l in list(self.p_axis.lines):
                l.remove()

            self.p_axis.plot(self.pressure_trace.zeroed_time*1000,
                             self.pressure_trace.raw_pressure,
                             'b', label="Raw Pressure")
            self.p_axis.plot(self.pressure_trace.zeroed_time*1000.0,
                             self.pressure_trace.pressure,
                             'g', label="Pressure")

        if hasattr(self, 'dpdt_axis'):
            for l in list(self.dpdt_axis.lines):
                l.remove()

            self.dpdt_axis.plot(self.pressure_trace.zeroed_time*1000.0,
                                self.pressure_trace.derivative/1000.0,
                                'm', label="Derivative")


    def change_filter_freq(self, value):
        """Change the cutoff frequency of the filter for the voltage trace

        Parameters
        ----------
        value : `float`
            The new value of the cutoff frequency
        """
        self.voltage_trace.change_filter_freq(value)
        self.pressure_trace = ExperimentalPressureTrace(self.voltage_trace,
                                                        self.experiment_parameters['pin'],
                                                        self.experiment_parameters['factor'],
                                                        )
        self.process_pressure_trace()
        self.copy_to_clipboard()
        if hasattr(self, 'p_axis'):
            # Create a list of the lines in the axis so that
            # removing one doesn't affect the list as its
            # looping. Use the loop rather than axis.clear()
            # so that the axis labels and legend are preserved.
            for l in list(self.p_axis.lines):
                l.remove()

            self.p_axis.plot(self.pressure_trace.zeroed_time*1000,
                             self.pressure_trace.raw_pressure,
                             'b', label="Raw Pressure")
            self.p_axis.plot(self.pressure_trace.zeroed_time*1000.0,
                             self.pressure_trace.pressure,
                             'g', label="Pressure")

        if hasattr(self, 'dpdt_axis'):
            for l in list(self.dpdt_axis.lines):
                l.remove()

            self.dpdt_axis.plot(self.pressure_trace.zeroed_time*1000.0,
                                self.pressure_trace.derivative/1000.0,
                                'm', label="Derivative")

    def copy_to_clipboard(self):
        """Copy experimental information to the clipboard

        The information is intended to be pasted to a spreadsheet.
        The information is ordered by columns in the following way:

        A. 4-digit time of day
        B. The initial pressure of the experiment, in Torr
        C. The initial temperature of the experiment, in K
        D. The end of compression pressure
        E. The overall ignition delay
        F. The first stage ignition delay
        G. The estimated end of compression temperature
        H. The inches of spacers for the experiment
        I. The millimeters of shims for the experiment
        J. The cutoff frequency that was used to filter the voltage trace
        """
        copy('\t'.join(map(str, [
            self.experiment_parameters['time_of_day'], self.experiment_parameters['pin'],
            self.experiment_parameters['Tin'], self.pressure_trace.p_EOC, self.ignition_delay,
            self.first_stage, self.T_EOC, self.experiment_parameters['spacers'],
            self.experiment_parameters['shims'], self.voltage_trace.filter_frequency])))

    def calculate_ignition_delay(self):
        """Calculate the ignition delay from the pressure trace.
        """
        # tau_points is an offset from the EOC to ensure that if
        # ignition is weak, the peak in dP/dt from the compression
        # stroke is not treated as the ignition event. Define points
        # to start looking for and stop looking for ignition.
        tau_points = int(0.002*self.pressure_trace.frequency)
        start_point = self.pressure_trace.EOC_idx + tau_points
        end_point = self.pressure_trace.EOC_idx + tau_points + 100000

        # The index of the maximum of the derivative trace
        # is taken as the point of ignition
        idx_of_ig = np.argmax(self.pressure_trace.derivative[start_point:end_point])

        # Take the offset into account when calculating the ignition
        # delay. The index of ignition is already relative to zero,
        # so we don't need to subtract any times. Stored in milliseconds
        ignition_delay = self.pressure_trace.time[idx_of_ig + tau_points]*1000

        try:
            end_first_stage = start_point + idx_of_ig - tau_points
            idx_of_first_stage = np.argmax(
                self.pressure_trace.derivative[start_point:end_first_stage]
            )
            first_stage = self.pressure_trace.time[idx_of_first_stage + tau_points]*1000
        except ValueError:
            first_stage = 0.0

        return ignition_delay, first_stage

    def calculate_EOC_temperature(self):
        est_comp_time = 0.03
        comp_time_length = int(est_comp_time*self.pressure_trace.frequency)
        pres_to_temp_start_idx = self.pressure_trace.EOC_idx - comp_time_length
        tempp = self.pressure_trace.pressure[pres_to_temp_start_idx:self.pressure_trace.EOC_idx]
        temperature_trace = TemperatureFromPressure(
            tempp,
            self.experiment_parameters['Tin'],
            chem_file=str(self.cti_file),
        )
        return np.amax(temperature_trace.temperature)

    def plot_pressure_trace(self):
        # Plot the filtered pressure and the derivative
        # on a new figure every time
        self.exp_fig = plt.figure(self.experiment_parameters['date'])
        self.p_axis = self.exp_fig.add_subplot(1, 1, 1)
        raw_line, = self.p_axis.plot(self.pressure_trace.zeroed_time*1000,
                                     self.pressure_trace.raw_pressure,
                                     'b', label="Raw Pressure")
        p_line, = self.p_axis.plot(self.pressure_trace.zeroed_time*1000.0,
                                   self.pressure_trace.pressure,
                                   'g', label="Pressure")
        self.dpdt_axis = self.p_axis.twinx()
        dpdt_line, = self.dpdt_axis.plot(self.pressure_trace.zeroed_time*1000.0,
                                         self.pressure_trace.derivative/1000.0,
                                         'm', label="Derivative")
        self.p_axis.legend([raw_line, p_line, dpdt_line],
                           [l.get_label() for l in [raw_line, p_line, dpdt_line]],
                           loc='best')
        self.p_axis.set_xlabel('Time [ms]')
        self.p_axis.set_ylabel('Pressure [bar]')
        self.dpdt_axis.set_ylabel('Time Derivative of Pressure [bar/ms]')
        if platform.system() == 'Windows':
            m = plt.get_current_fig_manager()
            m.window.showMaximized()


class AltExperiment(Experiment):
    """Contains all the information of a single alternate RCM experiment.
    See the documentation for `Experiment` for attribute descriptions.
    """
    def __init__(self, file_path=None, cti_file=None, cti_source=None, copy=True):
        self.resolve_file_path(file_path)
        self.experiment_parameters = self.parse_file_name(self.file_path)
        self.pressure_trace = AltExperimentalPressureTrace(self.file_path,
                                                           self.experiment_parameters['pin'])
        self.compression_time = None
        self.output_end_time = None
        self.offset_points = None
        if cti_source is None and cti_file is None:
            raise ValueError('One of cti_file or cti_source must be specified')
        elif cti_source is not None and cti_file is not None:
            raise ValueError('Only one of cti_file or cti_source can be specified')
        elif cti_source is None:
            self.cti_file = Path(cti_file).resolve()
            with open(str(cti_file), 'r') as in_file:
                self.cti_source = in_file.read()
        else:
            self.cti_source = cti_source
        self.process_pressure_trace()
        if copy:
            self.copy_to_clipboard()

    def __repr__(self):
        return 'AltExperiment(file_path={self.file_path!r})'.format(self=self)

    def copy_to_clipboard(self):
        """Copy experimental information to the clipboard

        The information is intended to be pasted to a spreadsheet.
        The information is ordered by columns in the following way:

        A. 4-digit time of day
        B. The initial pressure of the experiment, in Torr
        C. The initial temperature of the experiment, in K
        D. The end of compression pressure
        E. The overall ignition delay
        F. The first stage ignition delay
        G. The estimated end of compression temperature
        H. The inches of spacers for the experiment
        I. The millimeters of shims for the experiment
        J. The cutoff frequency that was used to filter the voltage trace
        """
        copy('\t'.join(map(str, [
            self.experiment_parameters['time_of_day'], self.experiment_parameters['pin'],
            self.experiment_parameters['Tin'], self.pressure_trace.p_EOC, self.ignition_delay,
            self.first_stage, self.T_EOC, self.experiment_parameters['spacers'],
            self.experiment_parameters['shims'], self.pressure_trace.filter_frequency])))

    def parse_file_name(self, file_path):
        """Parse the file name of an experimental trace, where the name is in an alternate format.

        Parameters
        ----------
        file_path : :class:`pathlib.Path`
            The Path object that contains the path of the current experimental data file.
            The filename associated with the path should be in the following format::

                'Fuel_FF_EqRatio_P.PP_PercentAr_AR_PercentN2_N2_[NR_]XX_in_YY_mm_ZZZK_AAAAtorr-DD-Mon-YY-Time_Endplug_CC.txt'

            where:

            - ``[NR_]``: Optional non-reactive indicator
            - ``FF``: Fuel symbol
            - ``P.PP``: Equivalence ratio to two decimal places
            - ``AR``: Percent of argon in the oxidizer
            - ``N2``: Percent of nitrogen in the oxidizer
            - ``XX``: inches of spacers
            - ``YY``: millimeters of shims
            - ``ZZZ``: Initial temperature in Kelvin
            - ``AAAA``: Initial pressure in Torr
            - ``BBB``: Multiplication factor set on the charge amplifier
            - ``DD-Mon-YY-Time``: Day, Month, Year, and Time of experiment
            - ``CC``: ``HC`` or ``LC`` to indicate the high clearance or
                      low clearance endplug, respectively

        Returns
        -------
        :class:`dict`
            Dictionary containing the parameters of the experiment with the
            following names:

            - ``spacers``: Inches of spacers
            - ``shims``: Millimeters of shims
            - ``Tin``: Initial temperature in Kelvin
            - ``pin``: Initial pressure in Torr
            - ``time_of_day``: Time of day of the experiment
            - ``date``: Date of the experiment with the time
            - ``date_year``: Date of the experiment with the year
        """
        name_parts = {}
        fname = file_path.name.rstrip('.txt')
        name_split = fname.split('_')
        try:
            name_split.remove('NR')
        except ValueError:
            pass
        name_split_end = name_split[13].split('-', maxsplit=1)
        data_date = datetime.strptime(name_split_end[1], '%d-%b-%y-%H%M')

        name_parts['spacers'] = int(name_split[8])/10
        name_parts['shims'] = int(name_split[10])
        name_parts['Tin'] = int(name_split[12][:-1])
        name_parts['pin'] = int(name_split_end[0].rstrip('torr'))
        name_parts['time_of_day'] = data_date.strftime('%H%M')
        name_parts['date'] = data_date.strftime('%d-%b-%H%M')
        name_parts['date_year'] = data_date.strftime('%d-%b-%y')
        return name_parts



"""
Module to test value of constants
"""

from ..constants import one_atm_in_torr, one_atm_in_bar, one_bar_in_pa


def test_one_atm_in_torr():
    assert one_atm_in_torr == 760.0


def test_one_atm_in_bar():
    assert one_atm_in_bar == 1.01325


def test_one_bar_in_pa():
    assert one_bar_in_pa == 1.0E5



"""
Test module for the experiments module
"""
import numpy as np
import os
import pytest
from ..experiments import Experiment


@pytest.fixture(scope='module')
def files():
    datadir = os.path.dirname(__file__)
    cti_file = os.path.join(datadir, 'species.cti')
    reacfile = os.path.join(datadir, '00_in_00_mm_333K-1146t-100x-21-Jul-15-1226.txt')
    nonrfile = os.path.join(datadir, 'NR_00_in_00_mm_333K-1137t-100x-21-Jul-15-1251.txt')
    return {'cti_file': cti_file, 'reacfile': reacfile, 'nonrfile': nonrfile}


def test_create_reactive_experiment(files):
    exp = Experiment(files['reacfile'],
                     cti_file=files['cti_file'], copy=False)
    assert np.isclose(exp.ignition_delay, 65.81)
    assert np.isclose(exp.T_EOC, 765.374)
    assert np.isclose(exp.first_stage, 63.80)
    assert np.isclose(exp.pressure_trace.p_EOC, 30.111)


def test_create_nonreactive_experiment(files):
    exp = Experiment(files['nonrfile'],
                     cti_file=files['cti_file'], copy=False)
    assert np.isclose(exp.ignition_delay, 0.0)
    assert np.isclose(exp.T_EOC, 0.0)
    assert np.isclose(exp.first_stage, 0.0)
    assert np.isclose(exp.pressure_trace.p_EOC, 30.184)



"""
Constants used in UConnRCMPy.
"""

one_atm_in_torr = 760.0
"""`float`: Conversion between atmosphere and Torr."""
one_atm_in_bar = 1.01325
"""`float`: Conversion between atmosphere and bar."""
one_bar_in_pa = 1.0E5
"""`float`: Conversion between bar and Pascal."""



from setuptools import setup, find_packages
import os.path as op
import sys

with open(op.join(op.dirname(op.realpath(__file__)), 'uconnrcmpy', '_version.py')) as version_file:
    exec(version_file.read())

with open(op.join(op.dirname(op.realpath(__file__)), 'README.md')) as readme_file:
    readme = readme_file.read()

with open(op.join(op.dirname(op.realpath(__file__)), 'CHANGELOG.md')) as changelog_file:
    changelog = changelog_file.read()

with open(op.join(op.dirname(op.realpath(__file__)), 'CITATION.md')) as citation_file:
    citation = citation_file.read()

desc = readme + '\n\n' + changelog + '\n\n' + citation
try:
    import pypandoc
    long_description = pypandoc.convert_text(desc, 'rst', format='md')
    with open(op.join(op.dirname(op.realpath(__file__)), 'README.rst'), 'w') as rst_readme:
        rst_readme.write(long_description)
except (ImportError, OSError, IOError):
    long_description = desc

install_requires = [
    'cantera>=2.3.0',
    'numpy>=1.8.0',
    'scipy>=0.18.0',
    'pyyaml>-3.12',
    'matplotlib>=1.4.0',
    'pyperclip>=1.5.27'
]

tests_require = [
    'pytest>=3.0.0',
    'pytest-cov>=2.3.1',
]

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
setup_requires = ['pytest-runner'] if needs_pytest else []

setup(
    name='UConnRCMPy',
    version=__version__,
    description='A package to process RCM data',
    long_description=long_description,
    url='https://github.com/bryanwweber/UConnRCMPy',
    author='Bryan W. Weber',
    author_email='bryan.weber@uconn.edu',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'processrcmfolder=uconnrcmpy.dataprocessing:process_folder',
        ],
    },
    install_requires=install_requires,
    tests_require=tests_require,
    setup_requires=setup_requires,
)



"""
Test module for the conditions module
"""
import pytest
from unittest import mock
import os
from ..conditions import Condition


class TestCondition(object):
    """
    """
    def test_create_condition_no_cti(self):
        cti_file = os.path.join(os.path.dirname(__file__), 'species.cti')
        with mock.patch('builtins.input', return_value=cti_file):
            Condition(plotting=False)

    def test_create_condition_with_cti(self):
        cti_file = os.path.join(os.path.dirname(__file__), 'species.cti')
        Condition(cti_file=cti_file, plotting=False)

    def test_add_experiment(self):
        datadir = os.path.dirname(__file__)
        cti_file = os.path.join(datadir, 'species.cti')
        reacfile = os.path.join(datadir, '00_in_00_mm_333K-1146t-100x-21-Jul-15-1226.txt')
        nonrfile = os.path.join(datadir, 'NR_00_in_00_mm_333K-1137t-100x-21-Jul-15-1251.txt')
        c = Condition(cti_file=cti_file, plotting=False)
        c.add_experiment(reacfile, copy=False)
        c.add_experiment(nonrfile, copy=False)



"""Data Processing Module"""

# System imports
from pathlib import Path
from glob import glob
import platform
import sys

# Third-party imports
import numpy as np
import matplotlib.pyplot as plt
import yaml
import cantera as ct

# Local imports
from pyperclip import copy
from .traces import (TemperatureFromPressure,
                     VolumeFromPressure,
                     PressureFromVolume,
                     )
from .experiments import Experiment, AltExperiment
from .simulations import Simulation


class Condition(object):
    """Class containing all the experiments at a condition.

    Parameters
    ----------
    cti_file : `str` or `pathlib.Path`, optional
        The location of the CTI file to use for Cantera. If it is
        not specified, the default is ``./species.cti``, but if that
        file cannot be found, the user is prompted for a filename.
    plotting : `bool`, optional
        Set to True to enable plotting when experiments are added

    Attributes
    ----------
    reactive_experiments : `dict`
        Dictionary of instances of class `Experiment` containing
        reactive experiments. Indexed by the file name of the
        underlying voltage trace from the experiment.
    nonreactive_experiments : `dict`
        Dictionary of instances of class `Experiment` containing
        nonreactive experiments. Indexed by the file name of the
        underlying voltage trace from the experiment.
    reactive_case : `Experiment`
        Single instance of class `Experiment` that is the closest to
        the mean of all of the experiments.
    nonreactive_case : `Experiment`
        Single instance of class `Experiment` that best matches the
        `~Condition.reactive_case`.
    presout : `numpy.ndarray`
        Pressure array that will be saved to the output file
    volout : `numpy.ndarray`
        Volume array that will be saved to the output file
    nonreactive_sim : `Simulation`
        Instance containing the nonreactive simulation
    reactive_sim : `Simulation`
        Instance containing the reactive simulation
    plotting : `bool`
        Set to True when plotting of experiments is enabled
    all_runs_figure : `matplotlib.figure.Figure`
        Figure showing all the runs at a condition
    nonreactive_figure : `matplotlib.figure.Figure`
        Figure showing a comparison of the nonreactive pressure trace
        with the reactive_case
    pressure_comparison_figure : `matplotlib.figure.Figure`
        Comparison figure of the reactive_case pressure and the
        pressure calculated by the volume trace routine
    simulation_figure : `matplotlib.figure.Figure`
        Comparison of the simulation with the reactive_case pressure
    """

    def __init__(self, cti_file=None, plotting=True):
        self.reactive_experiments = {}
        self.nonreactive_experiments = {}
        self.reactive_case = None
        self.reactive_file = None
        self.nonreactive_case = None
        self.nonreactive_file = None
        self.presout = None
        self.volout = None
        self.nonreactive_sim = None
        self.reactive_sim = None
        self.plotting = plotting
        if self.plotting:
            self.all_runs_figure = None
            self.all_runs_lines = {}
            self.nonreactive_figure = None
            self.pressure_comparison_figure = None
            self.simulation_figure = None

        self.output_attributes = [
            'reactive_file', 'nonreactive_file', 'nonreactive_end_time', 'reactive_end_time',
            'reactive_compression_time', 'nonreactive_offset_points', 'reactive_offset_points'
        ]
        if cti_file is None:
            path_args = {'strict': True} if sys.version_info >= (3, 6) else {}
            try:
                cti_file = str(Path('./species.cti').resolve(**path_args))
            except FileNotFoundError:
                cti_file = str(Path(input('Input the name of the CTI file: ')).resolve())

        self.cti_file = cti_file

        with open(str(cti_file), 'r') as in_file:
            self.cti_source = in_file.read()

        ct.suppress_thermo_warnings(False)
        ct.Solution(self.cti_file)
        ct.suppress_thermo_warnings()

    def __repr__(self):
        return 'Condition(cti_file, plotting={self.plotting!r})'.format(self=self)

    def summary(self):
        summary_str = [('Date-Time     P0 (Torr)  T0 (Kelvin)  Pc (bar)  τ (ms)  τ1 (ms)\n'
                        '---------     ---------  -----------  --------  ------  -------')]
        spacer = len(summary_str[0].split('\n')[0])*'-'
        summary_str.append('Reactive Experiments')
        summary_str.append(spacer)
        fmt = '{date}  {P0:^9d}  {T0:^11d}  {Pc:^8.2f}  {tau_ig:^6.2f}  {tau_first:^7.2f}'
        if not self.reactive_experiments:
            summary_str.append('No reactive experiments in this Condition.\n'
                               'Add some with add_experiment().')
        else:
            for key, expt in self.reactive_experiments.items():
                summary_str.append(fmt.format(
                    date=expt.experiment_parameters['date'],
                    P0=expt.experiment_parameters['pin'],
                    T0=expt.experiment_parameters['Tin'],
                    Pc=expt.pressure_trace.p_EOC,
                    tau_ig=expt.ignition_delay,
                    tau_first=expt.first_stage,
                ))
        summary_str.append(spacer)
        summary_str.append('Non-Reactive Experiments')
        summary_str.append(spacer)
        if not self.nonreactive_experiments:
            summary_str.append('No non-reactive experiments in this Condition.\n'
                               'Add some with add_experiment().')
        else:
            for key, expt in self.nonreactive_experiments.items():
                summary_str.append(fmt.format(
                    date=expt.experiment_parameters['date'],
                    P0=expt.experiment_parameters['pin'],
                    T0=expt.experiment_parameters['Tin'],
                    Pc=expt.pressure_trace.p_EOC,
                    tau_ig=expt.ignition_delay,
                    tau_first=expt.first_stage,
                ))
        return '\n'.join(summary_str)

    def __call__(self):
        print(self.summary())

    @property
    def reactive_file(self):
        """`pathlib.Path`: File name of the case that is the closest to the mean of all
        the experiments. Can be set with a `str` or `~pathlib.Path`.
        """
        return self._reactive_file

    @reactive_file.setter
    def reactive_file(self, value):
        if value is not None:
            self._reactive_file = Path(value).resolve()
        else:
            self._reactive_file = value

    @property
    def nonreactive_file(self):
        """`pathlib.Path`: File name of the non-reactive case that best matches
        the pressure trace of the `~Condition.reactive_case`. Can be set with a
        `str` or `~pathlib.Path`.
        """
        return self._nonreactive_file

    @nonreactive_file.setter
    def nonreactive_file(self, value):
        if value is not None:
            self._nonreactive_file = Path(value).resolve()
        else:
            self._nonreactive_file = value

    @property
    def reactive_end_time(self):
        """`float`: The end time for the output of the reactive case to a file, relative to
        the end of compression in milliseconds
        """
        if self.reactive_case is not None:
            return self.reactive_case.output_end_time
        else:
            raise AttributeError('The reactive_case has not been set for this Condition yet.')

    @reactive_end_time.setter
    def reactive_end_time(self, value):
        if self.reactive_case is not None:
            max_time = self.reactive_case.pressure_trace.zeroed_time[-1]*1000
            if value > max_time:
                raise ValueError('The reactive end time cannot be after the end of the pressure '
                                 'trace. The maximum time is: {} ms.'.format(max_time))

            self.reactive_case.output_end_time = value
        else:
            raise AttributeError('The reactive_case has not been set for this Condition yet.')

    @property
    def nonreactive_end_time(self):
        """`float`: The end time for the output of the nonreactive case to a file, relative to
        the end of compression in milliseconds
        """
        if self.nonreactive_case is not None:
            return self.nonreactive_case.output_end_time
        else:
            raise AttributeError('The nonreactive_case has not been set for this Condition yet.')

    @nonreactive_end_time.setter
    def nonreactive_end_time(self, value):
        if self.nonreactive_case is not None:
            max_time = self.nonreactive_case.pressure_trace.zeroed_time[-1]*1000
            if value > max_time:
                raise ValueError('The nonreactive end time cannot be after the end of the pressure'
                                 ' trace. The maximum time is: {} ms.'.format(max_time))

            self.nonreactive_case.output_end_time = value
        else:
            raise AttributeError('The nonreactive_case has not been set for this Condition yet.')

    @property
    def reactive_compression_time(self):
        """`float`: The compression time for the reactive case, in milliseconds
        """
        if self.reactive_case is not None:
            return self.reactive_case.compression_time
        else:
            raise AttributeError('The reactive_case has not been set for this Condition yet.')

    @reactive_compression_time.setter
    def reactive_compression_time(self, value):
        if self.reactive_case is not None:
            max_time = np.abs(self.reactive_case.pressure_trace.zeroed_time[0]*1000)
            if value > max_time:
                raise ValueError('The compression time cannot be longer than the time before'
                                 'compression. The maximum time is: {} ms'.format(max_time))

            self.reactive_case.compression_time = value
        else:
            raise AttributeError('The reactive_case has not been set for this Condition yet.')

    @property
    def reactive_offset_points(self):
        """`int`: The number of points to offset the end of compression, to better
        match the non-reactive and reactive pressure traces together"""
        if self.reactive_case is not None:
            return self.reactive_case.offset_points
        else:
            raise AttributeError('The reactive_case has not been set for this Condition yet.')

    @reactive_offset_points.setter
    def reactive_offset_points(self, value):
        if self.reactive_case is not None:
            self.reactive_case.offset_points = value
        else:
            raise AttributeError('The reactive_case has not been set for this Condition yet.')

    @property
    def nonreactive_offset_points(self):
        """`int`: The number of points to offset the end of compression, to better
        match the non-reactive and reactive pressure traces together"""
        if self.nonreactive_case is not None:
            return self.nonreactive_case.offset_points
        else:
            raise AttributeError('The nonreactive_case has not been set for this Condition yet.')

    @nonreactive_offset_points.setter
    def nonreactive_offset_points(self, value):
        if self.nonreactive_case is not None:
            self.nonreactive_case.offset_points = value
        else:
            raise AttributeError('The nonreactive_case has not been set for this Condition yet.')

    def add_reactive_case(self):
        """Add the `~Condition.reactive_case` to the `Condition` if
        it hasn't been set yet. If the `~Condition.reactive_case`
        exists but the name of the `~Condition.reactive_file` is
        different than the name of the existing case, the case is
        replaced. If neither of these is true, nothing is done.
        """
        if (self.reactive_case is None or
                self.reactive_case.file_path.name != self.reactive_file.name):
            if self.reactive_file in self.reactive_experiments:
                self.reactive_case = self.reactive_experiments[self.reactive_file]
            else:
                self.reactive_case = Experiment(self.reactive_file, cti_file=self.cti_file)
                self.reactive_experiments[self.reactive_file] = self.reactive_case

    def add_nonreactive_case(self):
        """Add the `~Condition.nonreactive_case` to the `Condition` if
        it hasn't been set yet. If the `~Condition.nonreactive_case`
        exists but the name of the `~Condition.nonreactive_file` is
        different than the name of the existing case, the case is
        replaced. If neither of these is true, nothing is done.
        """
        if (self.nonreactive_case is None or
                self.nonreactive_case.file_path.name != self.nonreactive_file.name):
            if self.nonreactive_file in self.nonreactive_experiments:
                self.nonreactive_case = self.nonreactive_experiments[self.nonreactive_file]
            else:
                self.nonreactive_case = Experiment(self.nonreactive_file, cti_file=self.cti_file)
                self.nonreactive_experiments[self.nonreactive_file] = self.nonreactive_case

    def add_experiment(self, file_name=None, **kwargs):
        """Add an experiment to the Condition.

        Parameters
        ----------
        file_name : `pathlib.Path` or `str` or `None`
            Filename of the file with the voltage trace of the
            experiment to be added.
        """
        exp = Experiment(file_name, cti_file=self.cti_file, **kwargs)
        if exp.pressure_trace.is_reactive:
            self.reactive_experiments[exp.file_path.name] = exp
            if self.plotting:
                self.plot_reactive_figures(exp)
        else:
            self.nonreactive_experiments[exp.file_path.name] = exp
            if self.plotting:
                self.plot_nonreactive_figure(exp)

    def change_EOC_time(self, experiment, time, is_reactive=True):
        """Change the EOC time manually for an experiment

        Parameters
        ----------
        experiment : `Experiment` or `str` or `pathlib.Path`
            The experiment to be modified, either as a filename
            or as the instance of an `Experiment` class.
        time : `float`
            The value for the EOC time
        is_reactive : `boolean`
            The experiment is reactive or not
        """
        if isinstance(experiment, (str, Path)):
            if str(experiment) in self.reactive_experiments:
                experiment = self.reactive_experiments[str(experiment)]
            elif str(experiment) in self.nonreactive_experiments:
                experiment = self.nonreactive_experiments[str(experiment)]
            else:
                raise ValueError('{} could not be found in the Condition. '
                                 'Did you add it with add_experiment?'.format(str(experiment)))

        experiment.change_EOC_time(time, is_reactive)
        if self.plotting:
            self.all_runs_lines[experiment.file_path.name].remove()
            self.all_runs_lines[experiment.file_path.name], = self.all_runs_axis.plot(
                experiment.pressure_trace.zeroed_time*1000.0,
                experiment.pressure_trace.pressure,
                label=experiment.experiment_parameters['date'],
            )
            self.all_runs_axis.legend(loc='best')

    def change_filter_freq(self, experiment, value):
        """Change the cutoff frequency of the filter for an experiment

        Parameters
        ----------
        experiment : `Experiment` or `str` or `pathlib.Path`
            The experiment to be modified, either as a filename
            or as the instance of an `Experiment` class.
        value : `float`
            The value for the cutoff frequency
        """
        if isinstance(experiment, (str, Path)):
            if str(experiment) in self.reactive_experiments:
                experiment = self.reactive_experiments[str(experiment)]
            elif str(experiment) in self.nonreactive_experiments:
                experiment = self.nonreactive_experiments[str(experiment)]
            else:
                raise ValueError('{} could not be found in the Condition. '
                                 'Did you add it with add_experiment?'.format(str(experiment)))

        experiment.change_filter_freq(value)
        if self.plotting:
            self.all_runs_lines[experiment.file_path.name].remove()
            self.all_runs_lines[experiment.file_path.name], = self.all_runs_axis.plot(
                experiment.pressure_trace.zeroed_time*1000.0,
                experiment.pressure_trace.pressure,
                label=experiment.experiment_parameters['date'],
            )
            self.all_runs_axis.legend(loc='best')

    def plot_reactive_figures(self, exp):
        """Plot the reactive pressure trace on figures.

        Plots the reactive pressure trace on two figures. The first is
        a comparison of all of the processed reactive pressure traces.
        The second plot is a comparison of the pressure trace and the
        derivative for a single experiment.

        Parameters
        ----------
        exp : `Experiment`
            Experiment to plot
        """
        # Plot the filtered pressure and overlay future runs
        if self.all_runs_figure is None:
            self.all_runs_figure = plt.figure('Reactive Pressure Trace Comparison')
            self.all_runs_axis = self.all_runs_figure.add_subplot(1, 1, 1)
            self.all_runs_axis.set_ylabel('Pressure [bar]')
            self.all_runs_axis.set_xlabel('Time [ms]')
            if platform.system() == 'Windows':
                m = plt.get_current_fig_manager()
                m.window.showMaximized()

        self.all_runs_lines[exp.file_path.name], = self.all_runs_axis.plot(
            exp.pressure_trace.zeroed_time*1000.0,
            exp.pressure_trace.pressure,
            label=exp.experiment_parameters['date'],
        )
        self.all_runs_axis.legend(loc='best')

        exp.plot_pressure_trace()

    def load_yaml(self):
        """
        Load the yaml file called ``volume-trace.yaml`` containing the
        information needed to construct the volume trace. This method
        reads the YAML file and sets the appropriate attributes of the
        `Condition` instance. All the following data are required unless
        otherwise noted. The format of the yaml file is::

            variable: value

        * ``nonreactive_file``: File name of the non-reactive pressure trace.
            Type: String
        * ``reactive_file``: File name of the reactive pressure trace.
            Type: String
        * ``reactive_compression_time``: Length of time of the compression stroke, in ms.
            Type: Float
        * ``nonreactive_end_time``: End time used for the produced volume trace, in ms.
            Type: Float
        * ``reactive_end_time``: End time of the output reactive pressure trace, in ms.
            Type: Float
        * ``reactive_offset_points``: Offset in number of points from EOC for the
            reactive case. Optional, defaults to zero.
            Type: Integer
        * ``nonreactive_offset_points``: Offset in number of points from EOC for the
            non-reactive case. Optional, defaults to zero.
            Type: Integer
        """
        with open('volume-trace.yaml') as yaml_file:
            yaml_data = yaml.safe_load(yaml_file)

        for attribute in ['nonreactive_file', 'reactive_file']:
            if attribute in yaml_data:
                setattr(self, attribute, yaml_data[attribute])

        self.add_reactive_case()
        self.add_nonreactive_case()

        for attribute in self.output_attributes:
            if attribute in yaml_data:
                setattr(self, attribute, yaml_data[attribute])

    def load_old_yaml(self):
        """
        Load the yaml file called ``volume-trace.yaml`` containing the
        information needed to construct the volume trace. All the
        following data are required unless otherwise noted. The format
        of the yaml file is::

            variable: value

        * ``nonrfile``: File name of the non-reactive pressure trace.
                        Type: String
        * ``reacfile``: File name of the reactive pressure trace.
                        Type: String
        * ``comptime``: Length of time of the compression stroke, in ms.
                        Type: Float
        * ``nonrend``: End time used for the produced volume trace, in ms.
                        Type: Float
        * ``reacend``: End time of the output reactive pressure trace, in ms.
                        Type: Float
        * ``reacoffs``: Offset in number of points from EOC for the
                        reactive case. Optional, defaults to zero.
                        Type: Integer
        * ``nonroffs``: Offset in number of points from EOC for the
                        non-reactive case. Optional, defaults to zero.
                        Type: Integer
        """
        with open('volume-trace.yaml') as yaml_file:
            yaml_data = yaml.safe_load(yaml_file)

        attributes = [('reacfile', 'reactive_file'),
                      ('nonrfile', 'nonreactive_file'),
                      ('nonrend', 'nonreactive_end_time'),
                      ('reacend', 'reactive_end_time'),
                      ('comptime', 'reactive_compression_time'),
                      ('nonroffs', 'nonreactive_offset_points'),
                      ('reacoffs', 'reactive_offset_points')]

        for yaml_key, attribute in attributes:
            if yaml_key in yaml_data:
                setattr(self, attribute, yaml_data[yaml_key])

    def plot_nonreactive_figure(self, exp):
        """Plot the nonreactive pressure traces on a figure.

        Plots the nonreactive pressure traces in comparison with the
        reactive pressure trace. Adds additional traces each time.

        Parameters
        ----------
        exp : `Experiment`
            Experiment to plot
        """
        if self.nonreactive_figure is None:
            self.nonreactive_figure = plt.figure('Non-Reactive Pressure Trace Comparison')
            self.nonreactive_axis = self.nonreactive_figure.add_subplot(1, 1, 1)
            self.nonreactive_axis.set_ylabel('Pressure [bar]')
            self.nonreactive_axis.set_xlabel('Time [ms]')
            if platform.system() == 'Windows':
                m = plt.get_current_fig_manager()
                m.window.showMaximized()

            if self.reactive_file is None:
                self.reactive_file = input('Reactive filename: ')

            self.add_reactive_case()

            self.nonreactive_axis.plot(
                self.reactive_case.pressure_trace.zeroed_time*1000.0,
                self.reactive_case.pressure_trace.pressure,
                label=self.reactive_case.experiment_parameters['date'],
            )

            linear_fit = self.reactive_case.pressure_trace.pressure_fit()
            reactive_line = np.polyval(linear_fit, self.reactive_case.pressure_trace.time)
            self.nonreactive_axis.plot(
                self.reactive_case.pressure_trace.zeroed_time*1000.0,
                reactive_line,
                label='Linear Fit to Initial Pressure',
            )

        self.nonreactive_axis.plot(
            exp.pressure_trace.zeroed_time*1000.0,
            exp.pressure_trace.pressure,
            label=exp.experiment_parameters['date'],
        )
        self.nonreactive_axis.legend(loc='best')

    def create_volume_trace(self):
        """
        Create the volume trace based on the information in the loaded
        yaml file.
        """
        if self.reactive_file is None:
            self.reactive_file = input('Reactive filename: ')

        self.add_reactive_case()

        if self.nonreactive_file is None:
            self.nonreactive_file = input('Non-Reactive filename: ')

        self.add_nonreactive_case()

        # Set the offsets to zero by default for the initial run through
        # The user should modify these after the first run if necessary
        for attribute in ['reactive_offset_points', 'nonreactive_offset_points']:
            if not hasattr(self, attribute) or getattr(self, attribute) is None:
                setattr(self, attribute, 0)

        for attribute in self.output_attributes:
            if not hasattr(self, attribute) or getattr(self, attribute) is None:
                temp_val = input('Specify a value for the {}: '.format(attribute))
                setattr(self, attribute, float(temp_val))

        nonreactive_end_idx = int(
            self.nonreactive_case.pressure_trace.EOC_idx + self.nonreactive_offset_points +
            self.nonreactive_end_time/1000.0*self.nonreactive_case.pressure_trace.frequency
        )

        reactive_end_idx = int(
            self.reactive_case.pressure_trace.EOC_idx + self.reactive_offset_points +
            self.reactive_end_time/1000.0*self.reactive_case.pressure_trace.frequency
        )

        reactive_start_point = int(
            self.reactive_case.pressure_trace.EOC_idx + self.reactive_offset_points -
            self.reactive_compression_time/1000.0*self.reactive_case.pressure_trace.frequency
        )

        stroke_pressure = self.reactive_case.pressure_trace.pressure[
            (reactive_start_point):(self.reactive_case.pressure_trace.EOC_idx +
                                    1 + self.reactive_offset_points)
        ]

        post_pressure = self.nonreactive_case.pressure_trace.pressure[
            (self.nonreactive_case.pressure_trace.EOC_idx + self.nonreactive_offset_points):(
                nonreactive_end_idx
            )
        ]

        print_pressure = self.reactive_case.pressure_trace.pressure[
            reactive_start_point:reactive_end_idx
        ]

        n_print_pts = len(print_pressure)
        time = np.arange(-self.reactive_compression_time/1000.0, self.nonreactive_end_time/1000.0,
                         1/self.reactive_case.pressure_trace.frequency)
        stroke_volume = VolumeFromPressure(
            stroke_pressure,
            1.0,
            self.reactive_case.experiment_parameters['Tin'],
            chem_file=str(self.cti_file),
        ).volume
        stroke_temperature = TemperatureFromPressure(
            stroke_pressure,
            self.reactive_case.experiment_parameters['Tin'],
            chem_file=str(self.cti_file),
        ).temperature

        post_volume = VolumeFromPressure(
            post_pressure,
            stroke_volume[-1],
            stroke_temperature[-1],
            chem_file=str(self.cti_file),
        ).volume

        # The post_volume array is indexed from the second element to
        # eliminate the duplicated element from the end of the stroke
        # volume array.
        volume = np.concatenate((stroke_volume, post_volume[1:]))

        computed_pressure = PressureFromVolume(
            volume[::5],
            stroke_pressure[0]*1E5,
            self.reactive_case.experiment_parameters['Tin'],
            chem_file=str(self.cti_file),
        ).pressure

        copy('{:.4f}'.format(stroke_pressure[0]))

        self.volout = np.vstack(
            (time[::5] + self.reactive_compression_time/1000, volume[::5])
        ).transpose()

        self.presout = np.vstack(
            (time[:n_print_pts:5] + self.reactive_compression_time/1000,
             print_pressure[::5])
        ).transpose()

        self.write_output(
            self.volout,
            self.presout,
            self.reactive_case.experiment_parameters['Tin'],
        )

        self.write_yaml()

        if self.plotting:
            if self.pressure_comparison_figure is None:
                self.pressure_comparison_figure = plt.figure('Pressure Trace Comparison')
                self.pressure_comparison_axis = self.pressure_comparison_figure.add_subplot(1, 1, 1)
                if platform.system() == 'Windows':
                    m = plt.get_current_fig_manager()
                    m.window.showMaximized()

            self.pressure_comparison_axis.cla()
            self.pressure_comparison_axis.set_ylabel("Pressure [bar]")
            self.pressure_comparison_axis.set_xlabel("Time [ms]")
            plot_time = (self.reactive_case.pressure_trace.zeroed_time -
                         self.reactive_offset_points/self.reactive_case.pressure_trace.frequency)
            self.pressure_comparison_axis.plot(
                plot_time*1000.0,
                self.reactive_case.pressure_trace.pressure,
                label="Reactive Pressure",
            )
            self.pressure_comparison_axis.plot(time[:n_print_pts:5]*1000.0, print_pressure[::5],
                                               label="Output Pressure")
            self.pressure_comparison_axis.plot(time[::5]*1000.0, computed_pressure,
                                               label="Computed Pressure")
            linear_fit = self.reactive_case.pressure_trace.pressure_fit(
                comptime=self.reactive_compression_time/1000,
            )
            reactive_line = np.polyval(linear_fit, self.reactive_case.pressure_trace.time)
            self.pressure_comparison_axis.plot(
                self.reactive_case.pressure_trace.zeroed_time*1000.0,
                reactive_line,
                label="Linear Fit to Initial Pressure",
            )
            self.pressure_comparison_axis.legend(loc='best')

    def write_yaml(self):
        """Write the volume-trace.yaml output file for storage of parameters.
        The YAML file format is detailed in `~Condition.load_yaml`.
        """

        yaml_data = {}
        for attribute in self.output_attributes:
            val = getattr(self, attribute)
            if isinstance(val, Path):
                yaml_data[attribute] = val.name
            else:
                yaml_data[attribute] = val

        with open('volume-trace.yaml', 'w') as yaml_file:
            yaml.dump(yaml_data, yaml_file, default_flow_style=False)

    def write_output(self, volout, presout, Tin):
        """
        Write the output files from the volume and pressure traces.
        The traces are sampled every 5 points to reduce the amount
        of required computational time. This has negligible effect on
        the computed pressure trace. The volume trace is written in
        ``csv`` format to the file ``volume.csv`` and the experimental
        pressure trace is written to the file
        ``Tc__P0__T0_XXXK_pressure.txt``, where ``XXX`` represents the
        initial temperature of the experiment. The user should fill the
        missing values into the file name after simulations are
        completed and the values are known.
        """
        np.savetxt('volume.csv', volout, delimiter=',')
        np.savetxt('Tc__P0__T0_{}K_pressure.txt'.format(Tin), presout, delimiter='\t')

    def run_simulation(self, run_reactive=False, run_nonreactive=True,
                       end_temp=2500.0, end_time=0.2):
        """Run the simulations for this condition.

        Parameters
        ----------
        run_reactive : `bool`, optional
            True to run the reactive case. False by default.
        run_nonreactive : `bool`, optional
            True to run the nonreactive case. True by default.
        end_temp : `float`, optional
            Temperature at which the simulation is ended
        end_time : `float`, optional
            Time at which the simulation is ended.
        """
        def process_choice(sim_type):
            choice = input('Are you sure you want to overwrite the {sim_type} simulation? '
                           'Input y or n: '.format(sim_type=sim_type))
            if choice.startswith('n'):
                return False
            elif choice.startswith('y'):
                return True
            else:
                raise IOError('Invalid input')

        if run_nonreactive:
            if self.nonreactive_sim is None:
                self.nonreactive_sim = Simulation(
                    initial_temperature=self.reactive_case.experiment_parameters['Tin'],
                    initial_pressure=self.presout[0, 1]*1E5,
                    volume=self.volout,
                    is_reactive=False,
                    end_temp=end_temp,
                    end_time=end_time,
                    chem_file=str(self.cti_file),
                )
            else:
                if process_choice('nonreactive'):
                    self.nonreactive_sim = Simulation(
                        initial_temperature=self.reactive_case.experiment_parameters['Tin'],
                        initial_pressure=self.presout[0, 1]*1E5,
                        volume=self.volout,
                        is_reactive=False,
                        end_temp=end_temp,
                        end_time=end_time,
                        chem_file=str(self.cti_file),
                    )
                else:
                    print('Nothing was done')

        if run_reactive:
            if self.reactive_sim is None:
                self.reactive_sim = Simulation(
                    initial_temperature=self.reactive_case.experiment_parameters['Tin'],
                    initial_pressure=self.presout[0, 1]*1E5,
                    volume=self.volout,
                    is_reactive=True,
                    end_temp=end_temp,
                    end_time=end_time,
                    chem_file=str(self.cti_file),
                )
            else:
                if process_choice('reactive'):
                    self.reactive_sim = Simulation(
                        initial_temperature=self.reactive_case.experiment_parameters['Tin'],
                        initial_pressure=self.presout[0, 1]*1E5,
                        volume=self.volout,
                        is_reactive=True,
                        end_temp=end_temp,
                        end_time=end_time,
                        chem_file=str(self.cti_file),
                    )
                else:
                    print('Nothing was done')

    def compare_to_sim(self, run_reactive=False, run_nonreactive=True):
        """Compare the experiments to the simulations.

        Run the simulations for this condition, and if plotting is on,
        generate the comparison plot.

        Parameters
        ----------
        run_reactive : `bool`, optional
            True to run the reactive comparison. False by default.
        run_nonreactive : `bool`, optional
            True to run the nonreactive comparison. True by default.
        """
        if self.presout is None:
            # Load the experimental pressure trace. Try the glob function first
            # and if it fails, ask the user for help.
            flist = glob('*pressure.txt')
            if not len(flist) == 1:
                flist = [input('Input the experimental pressure trace file name: ')]
            self.presout = np.genfromtxt(flist[0])

        self.run_simulation(run_reactive, run_nonreactive)

        # Plot the pressure traces together
        compression_time = self.reactive_compression_time

        if self.plotting:
            if self.simulation_figure is None:
                self.simulation_figure = plt.figure('Simulation Comparison')
                self.simulation_axis = self.simulation_figure.add_subplot(1, 1, 1)
                if platform.system() == 'Windows':
                    m = plt.get_current_fig_manager()
                    m.window.showMaximized()

            self.simulation_axis.cla()
            self.simulation_axis.set_xlabel('Time [ms]')
            self.simulation_axis.set_ylabel('Pressure [bar]')

            self.simulation_axis.plot(
                self.presout[:, 0]*1000 - compression_time,
                self.presout[:, 1],
                label="Experimental Pressure"
            )

            if self.nonreactive_sim is not None:
                self.simulation_axis.plot(
                    self.nonreactive_sim.time*1000 - compression_time,
                    self.nonreactive_sim.pressure,
                    label="Simulated Non-Reactive Pressure"
                )

            if self.reactive_sim is not None:
                self.simulation_axis.plot(
                    self.reactive_sim.time*1000 - compression_time,
                    self.reactive_sim.pressure,
                    label="Simulated Reactive Pressure"
                )
                der = self.reactive_sim.derivative
                self.simulation_axis.plot(
                    self.reactive_sim.time*1000 - compression_time,
                    der/np.amax(der)*np.amax(self.reactive_sim.pressure),
                    label="Scaled Derivative of Simulated Pressure"
                )

            self.simulation_axis.legend(loc='best')

        print_str = ''
        copy_str = ''

        if self.nonreactive_sim is not None:
            T_EOC = np.amax(self.nonreactive_sim.temperature)
            print_str += '{:.0f}'.format(T_EOC)
            copy_str += '{}'.format(T_EOC)

        if self.reactive_sim is not None:
            if self.nonreactive_sim is not None:
                print_str += '\t'
                copy_str += '\t\t\t\t'
            ignition_idx = np.argmax(self.reactive_sim.derivative)
            ignition_delay = self.reactive_sim.time[ignition_idx]*1000 - compression_time
            print_str += '{:.6f}'.format(ignition_delay)
            copy_str += '{}'.format(ignition_delay)

        if self.nonreactive_sim is not None and self.reactive_sim is not None:
            # Use np.where to find the index of EOC for reactive simulation
            # Use the given index to find T_EOC_reactive
            # Compare to T_EOC (nonreactive case) using np.isclose
            EOC_reactive_idx = np.where(self.reactive_sim.time >= compression_time/1000)
            T_EOC_reactive = self.reactive_sim.temperature[EOC_reactive_idx[0][0]]
            if not np.isclose(T_EOC, T_EOC_reactive, rtol=0.001):
                print('WARNING: Temperatures at Reactive EOC and non-reactive EOC did not match')
                print('T_EOC_nonreactive = {} K'.format(T_EOC))
                print('T_EOC_reactive = {} K'.format(T_EOC_reactive))
        print(print_str)
        copy(copy_str)


class AltCondition(Condition):
    """Class containing all of the alternate experiments at a condition
    """

    def add_experiment(self, file_name=None):
        """Add an experiment to the Condition.

        Parameters
        ----------
        file_name : `str` or `None`
            Filename of the file with the voltage trace of the
            experiment to be added.
        """
        exp = AltExperiment(file_name, cti_file=self.cti_file)
        if exp.pressure_trace.is_reactive:
            self.reactive_experiments[exp.file_path.name] = exp
            if self.plotting:
                self.plot_reactive_figures(exp)
        else:
            self.nonreactive_experiments[exp.file_path.name] = exp
            if self.plotting:
                self.plot_nonreactive_figure(exp)

    def add_reactive_case(self):
        """Add the `~AltCondition.reactive_case` to the `AltCondition` if
        it hasn't been set yet. If the `~AltCondition.reactive_case`
        exists but the name of the `~AltCondition.reactive_file` is
        different than the name of the existing case, the case is
        replaced. If neither of these is true, nothing is done.
        """
        if (self.reactive_case is None or
                self.reactive_case.file_path.name != self.reactive_file.name):
            if self.reactive_file in self.reactive_experiments:
                self.reactive_case = self.reactive_experiments[self.reactive_file]
            else:
                self.reactive_case = AltExperiment(self.reactive_file, cti_file=self.cti_file)
                self.reactive_experiments[self.reactive_file] = self.reactive_case

    def add_nonreactive_case(self):
        """Add the `~AltCondition.nonreactive_case` to the `AltCondition` if
        it hasn't been set yet. If the `~AltCondition.nonreactive_case`
        exists but the name of the `~AltCondition.nonreactive_file` is
        different than the name of the existing case, the case is
        replaced. If neither of these is true, nothing is done.
        """
        if (self.nonreactive_case is None or
                self.nonreactive_case.file_path.name != self.nonreactive_file.name):
            if self.nonreactive_file in self.nonreactive_experiments:
                self.nonreactive_case = self.nonreactive_experiments[self.nonreactive_file]
            else:
                self.nonreactive_case = AltExperiment(self.nonreactive_file, cti_file=self.cti_file)
                self.nonreactive_experiments[self.nonreactive_file] = self.nonreactive_case

    def __repr__(self):
        return 'AltCondition(plotting={!r})'.format(self.plotting)


def process_folder(cti_file, path='.', plot=False):
    """Process a folder of experimental files.

    Process a folder containing files with reactive experiments to
    calculate the ignition delays and copy a table with the results
    to the clipboard.

    Parameters
    ----------
    cti_file : `str` or `pathlib.Path`
        File containing the CTI for Cantera
    path : `str`, optional
        Path to folder to be analyzed. Defaults to the current folder.
    plot : `bool`, optional
        True to enable plotting. False by default.
    """
    p = Path(path)
    result = []

    cti_file = Path(cti_file).resolve()

    if plot:
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)

    for f in p.glob('[0-3]*.txt'):
        print(f)
        case = Experiment(f.resolve(), cti_file=cti_file, copy=False)
        result.append('\t'.join(map(str, [
            case.experiment_parameters['time_of_day'], case.experiment_parameters['pin'],
            case.experiment_parameters['Tin'], case.pressure_trace.p_EOC, case.ignition_delay,
            case.first_stage, case.T_EOC, case.experiment_parameters['spacers'],
            case.experiment_parameters['shims'], case.voltage_trace.filter_frequency])))
        if plot:
            ax.plot(case.pressure_trace.zeroed_time, case.pressure_trace.pressure,
                    label=case.experiment_parameters['date'])

    copy('\n'.join(sorted(result)))
    print('Finished')


def process_alt_folder(cti_file, path='.', plot=False):
    """Process a folder of alternative experimental files.

    Process a folder containing files with reactive experiments to
    calculate the ignition delays and copy a table with the results
    to the clipboard.

    Parameters
    ----------
    cti_file : `str` or `pathlib.Path`
        File containing the CTI for Cantera
    path : `str`, optional
        Path to folder to be analyzed. Defaults to the current folder.
    plot : `bool`, optional
        True to enable plotting. False by default.
    """
    p = Path(path)
    result = []

    cti_file = Path(cti_file).resolve()

    if plot:
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)

    for f in p.glob('Fuel_MF*.txt'):
        if 'NR' in f.name:
            continue
        print(f)
        case = AltExperiment(f.resolve(), cti_file=cti_file, copy=False)
        result.append('\t'.join(map(str, [
            case.experiment_parameters['date_year'], case.experiment_parameters['time_of_day'],
            case.experiment_parameters['pin'], case.experiment_parameters['Tin'],
            case.pressure_trace.p_EOC, case.ignition_delay, case.first_stage, case.T_EOC,
            case.experiment_parameters['spacers'], case.experiment_parameters['shims'],
            case.pressure_trace.filter_frequency, f.name])))
        if plot:
            ax.plot(case.pressure_trace.zeroed_time, case.pressure_trace.pressure,
                    label=case.experiment_parameters['date'])

    copy('\n'.join(sorted(result)))
    print('Finished')


if __name__ == '__main__':
    process_folder()
